.label {
    color: black;
}

table.citation td {
    float: left;
    min-width: 5%;
    max-width: 80%;
}

pre, dl.class, dl.type {
    box-shadow: 0px 1px 6px 1px lightgray;
}

dl.class, dl.type {
    padding: 5px;
}

.breathe-sectiondef {
    width: auto;
}

pre {
    overflow: auto;
    word-wrap: normal;
    white-space: pre;
}
